""":mod:`pokerkit.games` implements various poker game definitions."""

from __future__ import annotations

from abc import ABC
from collections.abc import Iterable, Mapping

from pokerkit.hands import (
    BadugiHand,
    EightOrBetterLowHand,
    OmahaEightOrBetterLowHand,
    OmahaHoldemHand,
    RegularLowHand,
    ShortDeckHoldemHand,
    StandardHighHand,
    StandardLowHand,
)
from pokerkit.state import BettingStructure, Opening, Automation, State, Street
from pokerkit.utilities import Deck, RankOrder


class Poker(ABC):
    """The abstract base class for poker games."""

    max_down_card_count: int
    """The maximum number of down cards."""
    max_up_card_count: int
    """The maximum number of up cards."""
    max_board_card_count: int
    """The maximum number of board cards."""
    rank_orders: tuple[RankOrder, ...]
    """The rank orders."""
    button_status: bool
    """The button status."""

    @classmethod
    def _clean_values(
            cls,
            values: int | Iterable[int] | Mapping[int, int] | None,
            count: int,
    ) -> tuple[int, ...]:
        if values is None:
            return (0,) * count
        elif isinstance(values, Mapping):
            parsed_values = [0] * count

            for key, value in values.items():
                parsed_values[key] = value

            return tuple(parsed_values)
        elif isinstance(values, Iterable):
            parsed_values = list(values)

            for i in range(len(parsed_values), count):
                parsed_values.append(0)

            return tuple(parsed_values)
        elif isinstance(values, int):
            return (values,) * count
        else:
            raise AssertionError


class TexasHoldem(Poker, ABC):
    """The abstract base class for Texas hold'em games."""

    max_down_card_count = 2
    max_up_card_count = 0
    max_board_card_count = 5
    rank_orders = (RankOrder.STANDARD,)
    button_status = True


class FixedLimitTexasHoldem(TexasHoldem):
    """The class for fixed-limit Texas hold'em games."""

    @classmethod
    def create_state(
            cls,
            automations: tuple[Automation, ...],
            ante_trimming_status: bool,
            antes: int | Iterable[int] | Mapping[int, int] | None,
            blinds_or_straddles: Iterable[int] | Mapping[int, int],
            small_bet: int,
            big_bet: int,
            starting_stacks: int | Iterable[int],
            player_count: int,
    ) -> State:
        """Create a fixed-limit Texas hold'em game.

        Below is an example hand in fixed-limit Texas hold'em.

        >>> state = FixedLimitTexasHoldem.create_state(
        ...     (
        ...         Automation.ANTE_POSTING,
        ...         Automation.BET_COLLECTION,
        ...         Automation.BLIND_OR_STRADDLE_POSTING,
        ...         Automation.CARD_BURNING,
        ...         Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
        ...         Automation.HAND_KILLING,
        ...         Automation.CHIPS_PUSHING,
        ...         Automation.CHIPS_PULLING,
        ...     ),
        ...     True,
        ...     None,
        ...     (1, 2),
        ...     2,
        ...     4,
        ...     200,
        ...     2,
        ... )

        Below shows the pre-flop dealings and actions.

        >>> state.deal_hole('AcAs')
        State.HoleDealing(player_index=0, cards=(Ac, As), statuses=(False, False))
        >>> state.deal_hole('7h6h')
        State.HoleDealing(player_index=1, cards=(7h, 6h), statuses=(False, False))

        >>> state.complete_bet_or_raise_to()
        State.CompletionBettingOrRaisingTo(player_index=1, amount=4)
        >>> state.complete_bet_or_raise_to()
        State.CompletionBettingOrRaisingTo(player_index=0, amount=6)
        >>> state.fold()
        State.Folding(player_index=1)

        Below show the final stacks.

        >>> state.stacks
        [204, 196]

        :param antes: The antes.
        :param blinds_or_straddles: The blinds or straddles.
        :param small_bet: The small bet.
        :param big_bet: The big bet.
        :param starting_stacks: The starting stacks.
        :param player_count: The number of players.
        :return: The created state.
        """
        return State(
            Deck.STANDARD,
            (StandardHighHand,),
            (
                Street(
                    False,
                    (False,) * 2,
                    0,
                    False,
                    Opening.POSITION,
                    small_bet,
                    4,
                ),
                Street(
                    True,
                    (),
                    3,
                    False,
                    Opening.POSITION,
                    small_bet,
                    4,
                ),
                Street(
                    True,
                    (),
                    1,
                    False,
                    Opening.POSITION,
                    big_bet,
                    4,
                ),
                Street(
                    True,
                    (),
                    1,
                    False,
                    Opening.POSITION,
                    big_bet,
                    4,
                ),
            ),
            BettingStructure.FIXED_LIMIT,
            automations,
            ante_trimming_status,
            cls._clean_values(antes, player_count),
            cls._clean_values(blinds_or_straddles, player_count),
            0,
            cls._clean_values(starting_stacks, player_count),
        )


class NoLimitTexasHoldem(TexasHoldem):
    """The class for no-limit Texas hold'em games."""

    @classmethod
    def create_state(
            cls,
            automations: tuple[Automation, ...],
            ante_trimming_status: bool,
            antes: int | Iterable[int] | Mapping[int, int] | None,
            blinds_or_straddles: Iterable[int] | Mapping[int, int],
            min_bet: int,
            starting_stacks: int | Iterable[int],
            player_count: int,
    ) -> State:
        """Create a no-limit Texas hold'em game.

        Below shows the first televised million dollar pot between
        Tom Dwan and Phil Ivey.

        Link: https://youtu.be/GnxFohpljqM

        >>> state = NoLimitTexasHoldem.create_state(
        ...     (
        ...         Automation.ANTE_POSTING,
        ...         Automation.BET_COLLECTION,
        ...         Automation.BLIND_OR_STRADDLE_POSTING,
        ...         Automation.CARD_BURNING,
        ...         Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
        ...         Automation.HAND_KILLING,
        ...         Automation.CHIPS_PUSHING,
        ...         Automation.CHIPS_PULLING,
        ...     ),
        ...     True,
        ...     500,
        ...     (1000, 2000),
        ...     2000,
        ...     (1125600, 2000000, 553500),
        ...     3,
        ... )

        Below shows the pre-flop dealings and actions.

        >>> state.deal_hole('Ac2d')  # Ivey
        State.HoleDealing(player_index=0, cards=(Ac, 2d), statuses=(False, False))
        >>> state.deal_hole('5h7s')  # Antonius*
        State.HoleDealing(player_index=1, cards=(5h, 7s), statuses=(False, False))
        >>> state.deal_hole('7h6h')  # Dwan
        State.HoleDealing(player_index=2, cards=(7h, 6h), statuses=(False, False))

        >>> state.complete_bet_or_raise_to(7000)  # Dwan
        State.CompletionBettingOrRaisingTo(player_index=2, amount=7000)
        >>> state.complete_bet_or_raise_to(23000)  # Ivey
        State.CompletionBettingOrRaisingTo(player_index=0, amount=23000)
        >>> state.fold()  # Antonius
        State.Folding(player_index=1)
        >>> state.check_or_call()  # Dwan
        State.CheckingOrCalling(player_index=2, amount=16000)

        Below shows the flop dealing and actions.

        >>> state.deal_board('Jc3d5c')
        State.BoardDealing(cards=(Jc, 3d, 5c))

        >>> state.complete_bet_or_raise_to(35000)  # Ivey
        State.CompletionBettingOrRaisingTo(player_index=0, amount=35000)
        >>> state.check_or_call()  # Dwan
        State.CheckingOrCalling(player_index=2, amount=35000)

        Below shows the turn dealing and actions.

        >>> state.deal_board('4h')
        State.BoardDealing(cards=(4h,))

        >>> state.complete_bet_or_raise_to(90000)  # Ivey
        State.CompletionBettingOrRaisingTo(player_index=0, amount=90000)
        >>> state.complete_bet_or_raise_to(232600)  # Dwan
        State.CompletionBettingOrRaisingTo(player_index=2, amount=232600)
        >>> state.complete_bet_or_raise_to(1067100)  # Ivey
        State.CompletionBettingOrRaisingTo(player_index=0, amount=1067100)
        >>> state.check_or_call()  # Dwan
        State.CheckingOrCalling(player_index=2, amount=262400)

        Below shows the river dealing.

        >>> state.deal_board('Jh')
        State.BoardDealing(cards=(Jh,))

        Below show the final stacks.

        >>> state.stacks
        [572100, 1997500, 1109500]

        :param antes: The antes.
        :param blinds_or_straddles: The blinds or straddles.
        :param min_bet: The min bet.
        :param starting_stacks: The starting stacks.
        :param player_count: The number of players.
        :return: The created state.
        """
        return State(
            Deck.STANDARD,
            (StandardHighHand,),
            (
                Street(
                    False,
                    (False,) * 2,
                    0,
                    False,
                    Opening.POSITION,
                    min_bet,
                    None,
                ),
                Street(
                    True,
                    (),
                    3,
                    False,
                    Opening.POSITION,
                    min_bet,
                    None,
                ),
                Street(
                    True,
                    (),
                    1,
                    False,
                    Opening.POSITION,
                    min_bet,
                    None,
                ),
                Street(
                    True,
                    (),
                    1,
                    False,
                    Opening.POSITION,
                    min_bet,
                    None,
                ),
            ),
            BettingStructure.NO_LIMIT,
            automations,
            ante_trimming_status,
            cls._clean_values(antes, player_count),
            cls._clean_values(blinds_or_straddles, player_count),
            0,
            cls._clean_values(starting_stacks, player_count),
        )


class NoLimitShortDeckHoldem(TexasHoldem):
    """The class for no-limit short-deck hold'em games."""

    max_down_card_count = 2
    max_up_card_count = 0
    max_board_card_count = 5
    rank_orders = (RankOrder.SHORT_DECK_HOLDEM,)
    button_status = True

    @classmethod
    def create_state(
            cls,
            automations: tuple[Automation, ...],
            ante_trimming_status: bool,
            antes: int | Iterable[int] | Mapping[int, int] | None,
            blinds_or_straddles: Iterable[int] | Mapping[int, int],
            min_bet: int,
            starting_stacks: int | Iterable[int],
            player_count: int,
    ) -> State:
        """Create a no-limit short-deck hold'em game.

        Below shows an all-in hand between Xuan and Phua.

        Link: https://youtu.be/QlgCcphLjaQ

        >>> state = NoLimitShortDeckHoldem.create_state(
        ...     (
        ...         Automation.ANTE_POSTING,
        ...         Automation.BET_COLLECTION,
        ...         Automation.BLIND_OR_STRADDLE_POSTING,
        ...         Automation.CARD_BURNING,
        ...         Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
        ...         Automation.HAND_KILLING,
        ...         Automation.CHIPS_PUSHING,
        ...         Automation.CHIPS_PULLING,
        ...     ),
        ...     True,
        ...     3000,
        ...     {-1: 3000},
        ...     3000,
        ...     (495000, 232000, 362000, 403000, 301000, 204000),
        ...     6,
        ... )

        Below shows the pre-flop dealings and actions.

        >>> state.deal_hole('Th8h')  # Badziakouski
        State.HoleDealing(player_index=0, cards=(Th, 8h), statuses=(False, False))
        >>> state.deal_hole('QsJd')  # Zhong
        State.HoleDealing(player_index=1, cards=(Qs, Jd), statuses=(False, False))
        >>> state.deal_hole('QhQd')  # Xuan
        State.HoleDealing(player_index=2, cards=(Qh, Qd), statuses=(False, False))
        >>> state.deal_hole('8d7c')  # Jun
        State.HoleDealing(player_index=3, cards=(8d, 7c), statuses=(False, False))
        >>> state.deal_hole('KhKs')  # Phua
        State.HoleDealing(player_index=4, cards=(Kh, Ks), statuses=(False, False))
        >>> state.deal_hole('8c7h')  # Koon
        State.HoleDealing(player_index=5, cards=(8c, 7h), statuses=(False, False))

        >>> state.check_or_call()  # Badziakouski
        State.CheckingOrCalling(player_index=0, amount=3000)
        >>> state.check_or_call()  # Zhong
        State.CheckingOrCalling(player_index=1, amount=3000)
        >>> state.complete_bet_or_raise_to(35000)  # Xuan
        State.CompletionBettingOrRaisingTo(player_index=2, amount=35000)
        >>> state.fold()  # Jun
        State.Folding(player_index=3)
        >>> state.complete_bet_or_raise_to(298000)  # Phua
        State.CompletionBettingOrRaisingTo(player_index=4, amount=298000)
        >>> state.fold()  # Koon
        State.Folding(player_index=5)
        >>> state.fold()  # Badziakouski
        State.Folding(player_index=0)
        >>> state.fold()  # Zhong
        State.Folding(player_index=1)
        >>> state.check_or_call()  # Xuan
        State.CheckingOrCalling(player_index=2, amount=263000)

        Below shows the flop dealing.

        >>> state.deal_board('9h6cKc')
        State.BoardDealing(cards=(9h, 6c, Kc))

        Below shows the turn dealing.

        >>> state.deal_board('Jh')
        State.BoardDealing(cards=(Jh,))

        Below shows the river dealing.

        >>> state.deal_board('Ts')
        State.BoardDealing(cards=(Ts,))

        Below show the final stacks.

        >>> state.stacks
        [489000, 226000, 684000, 400000, 0, 198000]

        :param antes: The antes.
        :param blinds_or_straddles: The blinds or straddles.
        :param min_bet: The min bet.
        :param starting_stacks: The starting stacks.
        :param player_count: The number of players.
        :return: The created state.
        """
        return State(
            Deck.SHORT_DECK_HOLDEM,
            (ShortDeckHoldemHand,),
            (
                Street(
                    False,
                    (False,) * 2,
                    0,
                    False,
                    Opening.POSITION,
                    min_bet,
                    None,
                ),
                Street(
                    True,
                    (),
                    3,
                    False,
                    Opening.POSITION,
                    min_bet,
                    None,
                ),
                Street(
                    True,
                    (),
                    1,
                    False,
                    Opening.POSITION,
                    min_bet,
                    None,
                ),
                Street(
                    True,
                    (),
                    1,
                    False,
                    Opening.POSITION,
                    min_bet,
                    None,
                ),
            ),
            BettingStructure.NO_LIMIT,
            automations,
            ante_trimming_status,
            cls._clean_values(antes, player_count),
            cls._clean_values(blinds_or_straddles, player_count),
            0,
            cls._clean_values(starting_stacks, player_count),
        )


class OmahaHoldem(Poker, ABC):
    """The abstract base class for Omaha hold'em games."""

    max_down_card_count = 4
    max_up_card_count = 0
    max_board_card_count = 5
    button_status = True


class PotLimitOmahaHoldem(OmahaHoldem):
    """The class for pot-limit Omaha hold'em games."""

    rank_orders = (RankOrder.STANDARD,)

    @classmethod
    def create_state(
            cls,
            automations: tuple[Automation, ...],
            ante_trimming_status: bool,
            antes: int | Iterable[int] | Mapping[int, int] | None,
            blinds_or_straddles: Iterable[int] | Mapping[int, int],
            min_bet: int,
            starting_stacks: int | Iterable[int],
            player_count: int,
    ) -> State:
        """Create a pot-limit Omaha hold'em game.

        Below shows the largest online poker pot every played between
        Patrik Antonius and Viktor Blom.

        Link: https://youtu.be/UMBm66Id2AA

        >>> state = PotLimitOmahaHoldem.create_state(
        ...     (
        ...         Automation.ANTE_POSTING,
        ...         Automation.BET_COLLECTION,
        ...         Automation.BLIND_OR_STRADDLE_POSTING,
        ...         Automation.CARD_BURNING,
        ...         Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
        ...         Automation.HAND_KILLING,
        ...         Automation.CHIPS_PUSHING,
        ...         Automation.CHIPS_PULLING,
        ...     ),
        ...     True,
        ...     None,
        ...     (50000, 100000),
        ...     2000,
        ...     (125945025, 67847350),
        ...     2,
        ... )

        Below shows the pre-flop dealings and actions.

        >>> state.deal_hole('Ah3sKsKh')  # Antonius
        State.HoleDealing(player_index=0, cards=(Ah, 3s, Ks, Kh), statuses=(False, False, False, False))
        >>> state.deal_hole('6d9s7d8h')  # Blom
        State.HoleDealing(player_index=1, cards=(6d, 9s, 7d, 8h), statuses=(False, False, False, False))

        >>> state.complete_bet_or_raise_to(300000)  # Blom
        State.CompletionBettingOrRaisingTo(player_index=1, amount=300000)
        >>> state.complete_bet_or_raise_to(900000)  # Antonius
        State.CompletionBettingOrRaisingTo(player_index=0, amount=900000)
        >>> state.complete_bet_or_raise_to(2700000)  # Blom
        State.CompletionBettingOrRaisingTo(player_index=1, amount=2700000)
        >>> state.complete_bet_or_raise_to(8100000)  # Antonius
        State.CompletionBettingOrRaisingTo(player_index=0, amount=8100000)
        >>> state.check_or_call()  # Blom
        State.CheckingOrCalling(player_index=1, amount=5400000)

        Below shows the flop dealing and actions.

        >>> state.deal_board('4s5c2h')
        State.BoardDealing(cards=(4s, 5c, 2h))

        >>> state.complete_bet_or_raise_to(9100000)  # Antonius
        State.CompletionBettingOrRaisingTo(player_index=0, amount=9100000)
        >>> state.complete_bet_or_raise_to(43500000)  # Blom
        State.CompletionBettingOrRaisingTo(player_index=1, amount=43500000)
        >>> state.complete_bet_or_raise_to(77900000)  # Antonius
        State.CompletionBettingOrRaisingTo(player_index=0, amount=77900000)
        >>> state.check_or_call()  # Blom
        State.CheckingOrCalling(player_index=1, amount=16247350)

        Below shows the turn dealing.

        >>> state.deal_board('5h')
        State.BoardDealing(cards=(5h,))

        Below shows the river dealing.

        >>> state.deal_board('9c')
        State.BoardDealing(cards=(9c,))

        Below show the final stacks.

        >>> state.stacks
        [193792375, 0]

        :param antes: The antes.
        :param blinds_or_straddles: The blinds or straddles.
        :param min_bet: The min bet.
        :param starting_stacks: The starting stacks.
        :param player_count: The number of players.
        :return: The created state.
        """
        return State(
            Deck.STANDARD,
            (OmahaHoldemHand,),
            (
                Street(
                    False,
                    (False,) * 4,
                    0,
                    False,
                    Opening.POSITION,
                    min_bet,
                    None,
                ),
                Street(
                    True,
                    (),
                    3,
                    False,
                    Opening.POSITION,
                    min_bet,
                    None,
                ),
                Street(
                    True,
                    (),
                    1,
                    False,
                    Opening.POSITION,
                    min_bet,
                    None,
                ),
                Street(
                    True,
                    (),
                    1,
                    False,
                    Opening.POSITION,
                    min_bet,
                    None,
                ),
            ),
            BettingStructure.POT_LIMIT,
            automations,
            ante_trimming_status,
            cls._clean_values(antes, player_count),
            cls._clean_values(blinds_or_straddles, player_count),
            0,
            cls._clean_values(starting_stacks, player_count),
        )


class FixedLimitOmahaHoldemHighLowSplitEightOrBetter(OmahaHoldem):
    """The class for fixed-limit Omaha hold'em high/low-split eight or
    better low games.
    """

    rank_orders = RankOrder.STANDARD, RankOrder.EIGHT_OR_BETTER_LOW

    @classmethod
    def create_state(
            cls,
            automations: tuple[Automation, ...],
            ante_trimming_status: bool,
            antes: int | Iterable[int] | Mapping[int, int] | None,
            blinds_or_straddles: Iterable[int] | Mapping[int, int],
            small_bet: int,
            big_bet: int,
            starting_stacks: int | Iterable[int],
            player_count: int,
    ) -> State:
        """Create a fixed-limit Omaha hold'em high/low-split eight or better
        low game.

        :param antes: The antes.
        :param blinds_or_straddles: The blinds or straddles.
        :param small_bet: The small bet.
        :param big_bet: The big bet.
        :param starting_stacks: The starting stacks.
        :param player_count: The number of players.
        :return: The created state.
        """
        return State(
            Deck.STANDARD,
            (OmahaHoldemHand, OmahaEightOrBetterLowHand),
            (
                Street(
                    False,
                    (False,) * 4,
                    0,
                    False,
                    Opening.POSITION,
                    small_bet,
                    4,
                ),
                Street(
                    True,
                    (),
                    3,
                    False,
                    Opening.POSITION,
                    small_bet,
                    4,
                ),
                Street(
                    True,
                    (),
                    1,
                    False,
                    Opening.POSITION,
                    big_bet,
                    4,
                ),
                Street(
                    True,
                    (),
                    1,
                    False,
                    Opening.POSITION,
                    big_bet,
                    4,
                ),
            ),
            BettingStructure.FIXED_LIMIT,
            automations,
            ante_trimming_status,
            cls._clean_values(antes, player_count),
            cls._clean_values(blinds_or_straddles, player_count),
            0,
            cls._clean_values(starting_stacks, player_count),
        )


class SevenCardStud(Poker, ABC):
    """The abstract base class for seven card stud games."""

    max_down_card_count = 3
    max_up_card_count = 4
    max_board_card_count = 0
    button_status = False


class FixedLimitSevenCardStud(SevenCardStud):
    """The class for fixed-limit seven card stud games."""

    rank_orders = (RankOrder.STANDARD,)

    @classmethod
    def create_state(
            cls,
            automations: tuple[Automation, ...],
            ante_trimming_status: bool,
            antes: int | Iterable[int] | Mapping[int, int] | None,
            bring_in: int,
            small_bet: int,
            big_bet: int,
            starting_stacks: int | Iterable[int],
            player_count: int,
    ) -> State:
        """Create a fixed-limit seven card stud game.

        :param antes: The antes.
        :param bring_in: The bring-in.
        :param small_bet: The small bet.
        :param big_bet: The big bet.
        :param starting_stacks: The starting stacks.
        :param player_count: The number of players.
        :return: The created state.
        """
        return State(
            Deck.STANDARD,
            (StandardHighHand,),
            (
                Street(
                    False,
                    (False, False, True),
                    0,
                    False,
                    Opening.LOW_CARD,
                    small_bet,
                    4,
                ),
                Street(
                    True, (True,),
                    0,
                    False,
                    Opening.HIGH_HAND,
                    small_bet,
                    4,
                ),
                Street(
                    True,
                    (True,),
                    0,
                    False,
                    Opening.HIGH_HAND,
                    big_bet,
                    4,
                ),
                Street(
                    True,
                    (True,),
                    0,
                    False,
                    Opening.HIGH_HAND,
                    big_bet,
                    4,
                ),
                Street(
                    True,
                    (False,),
                    0,
                    False,
                    Opening.HIGH_HAND,
                    big_bet,
                    4,
                ),
            ),
            BettingStructure.FIXED_LIMIT,
            automations,
            ante_trimming_status,
            cls._clean_values(antes, player_count),
            cls._clean_values(0, player_count),
            bring_in,
            cls._clean_values(starting_stacks, player_count),
        )


class FixedLimitSevenCardStudHighLowSplitEightOrBetter(SevenCardStud):
    """The class for fixed-limit seven card stud high/low-split eight or
    better low games.
    """

    rank_orders = RankOrder.STANDARD, RankOrder.EIGHT_OR_BETTER_LOW

    @classmethod
    def create_state(
            cls,
            automations: tuple[Automation, ...],
            ante_trimming_status: bool,
            antes: int | Iterable[int] | Mapping[int, int] | None,
            bring_in: int,
            small_bet: int,
            big_bet: int,
            starting_stacks: int | Iterable[int],
            player_count: int,
    ) -> State:
        """Create a fixed-limit seven card stud high/low-split eight or
        better low game.

        :param antes: The antes.
        :param bring_in: The bring-in.
        :param small_bet: The small bet.
        :param big_bet: The big bet.
        :param starting_stacks: The starting stacks.
        :param player_count: The number of players.
        :return: The created state.
        """
        return State(
            Deck.STANDARD,
            (StandardHighHand, EightOrBetterLowHand),
            (
                Street(
                    False,
                    (False, False, True),
                    0,
                    False,
                    Opening.LOW_CARD,
                    small_bet,
                    4,
                ),
                Street(
                    True,
                    (True,),
                    0,
                    False,
                    Opening.HIGH_HAND,
                    small_bet,
                    4,
                ),
                Street(
                    True,
                    (True,),
                    0,
                    False,
                    Opening.HIGH_HAND,
                    big_bet,
                    4,
                ),
                Street(
                    True,
                    (True,),
                    0,
                    False,
                    Opening.HIGH_HAND,
                    big_bet,
                    4,
                ),
                Street(
                    True,
                    (False,),
                    0,
                    False,
                    Opening.HIGH_HAND,
                    big_bet,
                    4,
                ),
            ),
            BettingStructure.FIXED_LIMIT,
            automations,
            ante_trimming_status,
            cls._clean_values(antes, player_count),
            cls._clean_values(0, player_count),
            bring_in,
            cls._clean_values(starting_stacks, player_count),
        )


class FixedLimitRazz(SevenCardStud):
    """The class for fixed-limit razz games."""

    rank_orders = (RankOrder.REGULAR,)

    @classmethod
    def create_state(
            cls,
            automations: tuple[Automation, ...],
            ante_trimming_status: bool,
            antes: int | Iterable[int] | Mapping[int, int] | None,
            bring_in: int,
            small_bet: int,
            big_bet: int,
            starting_stacks: int | Iterable[int],
            player_count: int,
    ) -> State:
        """Create a fixed-limit razz game.

        :param antes: The antes.
        :param bring_in: The bring-in.
        :param small_bet: The small bet.
        :param big_bet: The big bet.
        :param starting_stacks: The starting stacks.
        :param player_count: The number of players.
        :return: The created state.
        """
        return State(
            Deck.REGULAR,
            (RegularLowHand,),
            (
                Street(
                    False,
                    (False, False, True),
                    0,
                    False,
                    Opening.HIGH_CARD,
                    small_bet,
                    4,
                ),
                Street(
                    True,
                    (True,),
                    0,
                    False,
                    Opening.LOW_HAND,
                    small_bet,
                    4,
                ),
                Street(
                    True,
                    (True,),
                    0,
                    False,
                    Opening.LOW_HAND,
                    big_bet,
                    4,
                ),
                Street(
                    True,
                    (True,),
                    0,
                    False,
                    Opening.LOW_HAND,
                    big_bet,
                    4,
                ),
                Street(
                    True,
                    (False,),
                    0,
                    False,
                    Opening.LOW_HAND,
                    big_bet,
                    4,
                ),
            ),
            BettingStructure.FIXED_LIMIT,
            automations,
            ante_trimming_status,
            cls._clean_values(antes, player_count),
            cls._clean_values(0, player_count),
            bring_in,
            cls._clean_values(starting_stacks, player_count),
        )


class DeuceToSevenLowball(Poker, ABC):
    """The abstract base class for deuce-to-seven lowball games."""

    max_down_card_count = 5
    max_up_card_count = 0
    max_board_card_count = 0
    rank_orders = (RankOrder.STANDARD,)
    button_status = True


class NoLimitDeuceToSevenLowballSingleDraw(DeuceToSevenLowball):
    """The class for no-limit deuce-to-seven lowball single draw games.
    """

    @classmethod
    def create_state(
            cls,
            automations: tuple[Automation, ...],
            ante_trimming_status: bool,
            antes: int | Iterable[int] | Mapping[int, int] | None,
            blinds_or_straddles: Iterable[int] | Mapping[int, int],
            min_bet: int,
            starting_stacks: int | Iterable[int],
            player_count: int,
    ) -> State:
        """Create a no-limit deuce-to-seven lowball single draw game.

        :param antes: The antes.
        :param blinds_or_straddles: The blinds or straddles.
        :param min_bet: The min bet.
        :param starting_stacks: The starting stacks.
        :param player_count: The number of players.
        :return: The created state.
        """
        return State(
            Deck.STANDARD,
            (StandardLowHand,),
            (
                Street(
                    False,
                    (False,) * 5,
                    0,
                    False,
                    Opening.POSITION,
                    min_bet,
                    None,
                ),
                Street(
                    True,
                    (),
                    0,
                    True,
                    Opening.POSITION,
                    min_bet,
                    None,
                ),
            ),
            BettingStructure.NO_LIMIT,
            automations,
            ante_trimming_status,
            cls._clean_values(antes, player_count),
            cls._clean_values(blinds_or_straddles, player_count),
            0,
            cls._clean_values(starting_stacks, player_count),
        )


class FixedLimitDeuceToSevenLowballTripleDraw(DeuceToSevenLowball):
    """The class for fixed-limit deuce-to-seven lowball triple draw
    games.
    """

    @classmethod
    def create_state(
            cls,
            automations: tuple[Automation, ...],
            ante_trimming_status: bool,
            antes: int | Iterable[int] | Mapping[int, int] | None,
            blinds_or_straddles: Iterable[int] | Mapping[int, int],
            small_bet: int,
            big_bet: int,
            starting_stacks: int | Iterable[int],
            player_count: int,
    ) -> State:
        """Create a fixed-limit deuce-to-seven lowball triple draw game.

        Below shows a bad beat between Yockey and Arieh.

        Link: https://youtu.be/pChCqb2FNxY

        >>> state = FixedLimitDeuceToSevenLowballTripleDraw.create_state(
        ...     (
        ...         Automation.ANTE_POSTING,
        ...         Automation.BET_COLLECTION,
        ...         Automation.BLIND_OR_STRADDLE_POSTING,
        ...         Automation.CARD_BURNING,
        ...         Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
        ...         Automation.HAND_KILLING,
        ...         Automation.CHIPS_PUSHING,
        ...         Automation.CHIPS_PULLING,
        ...     ),
        ...     True,
        ...     None,
        ...     (75000, 150000),
        ...     150000,
        ...     300000,
        ...     (1180000, 4340000, 5910000, 10765000),
        ...     4,
        ... )

        Below shows the pre-flop dealings and actions.

        >>> state.deal_hole('7h6c4c3d2c')  # Yockey
        State.HoleDealing(player_index=0, cards=(7h, 6c, 4c, 3d, 2c), statuses=(False, False, False, False, False))
        >>> state.deal_hole('JsJcJdJhTs')  # Hui*
        State.HoleDealing(player_index=1, cards=(Js, Jc, Jd, Jh, Ts), statuses=(False, False, False, False, False))
        >>> state.deal_hole('KsKcKdKhTh')  # Esposito*
        State.HoleDealing(player_index=2, cards=(Ks, Kc, Kd, Kh, Th), statuses=(False, False, False, False, False))
        >>> state.deal_hole('AsQs6s5c3c')  # Arieh
        State.HoleDealing(player_index=3, cards=(As, Qs, 6s, 5c, 3c), statuses=(False, False, False, False, False))

        >>> state.fold()  # Esposito
        State.Folding(player_index=2)
        >>> state.complete_bet_or_raise_to()  # Arieh
        State.CompletionBettingOrRaisingTo(player_index=3, amount=300000)
        >>> state.complete_bet_or_raise_to()  # Yockey
        State.CompletionBettingOrRaisingTo(player_index=0, amount=450000)
        >>> state.fold()  # Hui
        State.Folding(player_index=1)
        >>> state.check_or_call()  # Arieh
        State.CheckingOrCalling(player_index=3, amount=150000)

        Below shows the first draw and actions.

        >>> state.stand_pat_or_discard()  # Yockey
        State.StandingPatOrDiscarding(player_index=0, cards=())
        >>> state.stand_pat_or_discard('AsQs')  # Arieh
        State.StandingPatOrDiscarding(player_index=3, cards=(As, Qs))
        >>> state.deal_hole('2hQh')  # Arieh
        State.HoleDealing(player_index=3, cards=(2h, Qh), statuses=(False, False))

        >>> state.complete_bet_or_raise_to()  # Yockey
        State.CompletionBettingOrRaisingTo(player_index=0, amount=150000)
        >>> state.check_or_call()  # Arieh
        State.CheckingOrCalling(player_index=3, amount=150000)

        Below shows the second draw and actions.

        >>> state.stand_pat_or_discard()  # Yockey
        State.StandingPatOrDiscarding(player_index=0, cards=())
        >>> state.stand_pat_or_discard('Qh')  # Arieh
        State.StandingPatOrDiscarding(player_index=3, cards=(Qh,))
        >>> state.deal_hole('4d')  # Arieh
        State.HoleDealing(player_index=3, cards=(4d,), statuses=(False,))

        >>> state.complete_bet_or_raise_to()  # Yockey
        State.CompletionBettingOrRaisingTo(player_index=0, amount=300000)
        >>> state.check_or_call()  # Arieh
        State.CheckingOrCalling(player_index=3, amount=300000)

        Below shows the third draw and actions.

        >>> state.stand_pat_or_discard()  # Yockey
        State.StandingPatOrDiscarding(player_index=0, cards=())
        >>> state.stand_pat_or_discard('6s')  # Arieh
        State.StandingPatOrDiscarding(player_index=3, cards=(6s,))
        >>> state.deal_hole('7c')  # Arieh
        State.HoleDealing(player_index=3, cards=(7c,), statuses=(False,))

        >>> state.complete_bet_or_raise_to()  # Yockey
        State.CompletionBettingOrRaisingTo(player_index=0, amount=280000)
        >>> state.check_or_call()  # Arieh
        State.CheckingOrCalling(player_index=3, amount=280000)

        Below show the final stacks.

        >>> state.stacks
        [0, 4190000, 5910000, 12095000]

        :param antes: The antes.
        :param blinds_or_straddles: The blinds or straddles.
        :param small_bet: The small bet.
        :param big_bet: The big bet.
        :param starting_stacks: The starting stacks.
        :param player_count: The number of players.
        :return: The created state.
        """
        return State(
            Deck.STANDARD,
            (StandardLowHand,),
            (
                Street(
                    False,
                    (False,) * 5,
                    0,
                    False,
                    Opening.POSITION,
                    small_bet,
                    None,
                ),
                Street(
                    True,
                    (),
                    0,
                    True,
                    Opening.POSITION,
                    small_bet,
                    None,
                ),
                Street(
                    True,
                    (),
                    0,
                    True,
                    Opening.POSITION,
                    big_bet,
                    None,
                ),
                Street(
                    True,
                    (),
                    0,
                    True,
                    Opening.POSITION,
                    big_bet,
                    None,
                ),
            ),
            BettingStructure.FIXED_LIMIT,
            automations,
            ante_trimming_status,
            cls._clean_values(antes, player_count),
            cls._clean_values(blinds_or_straddles, player_count),
            0,
            cls._clean_values(starting_stacks, player_count),
        )


class FixedLimitBadugi(Poker):
    """The class for fixed-limit badugi games."""

    max_down_card_count = 4
    max_up_card_count = 0
    max_board_card_count = 0
    rank_orders = (RankOrder.REGULAR,)
    button_status = True

    @classmethod
    def create_state(
            cls,
            automations: tuple[Automation, ...],
            ante_trimming_status: bool,
            antes: int | Iterable[int] | Mapping[int, int] | None,
            blinds_or_straddles: Iterable[int] | Mapping[int, int],
            small_bet: int,
            big_bet: int,
            starting_stacks: int | Iterable[int],
            player_count: int,
    ) -> State:
        """Create a fixed-limit badugi game.

        Below shows an example badugi hand from Wikipedia.

        Link: https://en.wikipedia.org/wiki/Badugi

        >>> state = FixedLimitBadugi.create_state(
        ...     (
        ...         Automation.ANTE_POSTING,
        ...         Automation.BET_COLLECTION,
        ...         Automation.BLIND_OR_STRADDLE_POSTING,
        ...         Automation.CARD_BURNING,
        ...         Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
        ...         Automation.HAND_KILLING,
        ...         Automation.CHIPS_PUSHING,
        ...         Automation.CHIPS_PULLING,
        ...     ),
        ...     True,
        ...     None,
        ...     (1, 2),
        ...     2,
        ...     4,
        ...     200,
        ...     4,
        ... )

        Below shows the pre-flop dealings and actions.

        >>> state.deal_hole('As4hJcKh')  # Bob*
        State.HoleDealing(player_index=0, cards=(As, 4h, Jc, Kh), statuses=(False, False, False, False))
        >>> state.deal_hole('3s5d7s8s')  # Carol*
        State.HoleDealing(player_index=1, cards=(3s, 5d, 7s, 8s), statuses=(False, False, False, False))
        >>> state.deal_hole('KsKdQsQd')  # Ted*
        State.HoleDealing(player_index=2, cards=(Ks, Kd, Qs, Qd), statuses=(False, False, False, False))
        >>> state.deal_hole('2s4c6dKc')  # Alice*
        State.HoleDealing(player_index=3, cards=(2s, 4c, 6d, Kc), statuses=(False, False, False, False))

        >>> state.fold()  # Ted
        State.Folding(player_index=2)
        >>> state.check_or_call()  # Alice
        State.CheckingOrCalling(player_index=3, amount=2)
        >>> state.check_or_call()  # Bob
        State.CheckingOrCalling(player_index=0, amount=1)
        >>> state.check_or_call()  # Carol
        State.CheckingOrCalling(player_index=1, amount=0)

        Below shows the first draw and actions.

        >>> state.stand_pat_or_discard('JcKh')  # Bob*
        State.StandingPatOrDiscarding(player_index=0, cards=(Jc, Kh))
        >>> state.stand_pat_or_discard('7s8s')  # Carol*
        State.StandingPatOrDiscarding(player_index=1, cards=(7s, 8s))
        >>> state.stand_pat_or_discard('Kc')  # Alice*
        State.StandingPatOrDiscarding(player_index=3, cards=(Kc,))
        >>> state.deal_hole('TcJs')  # Bob*
        State.HoleDealing(player_index=0, cards=(Tc, Js), statuses=(False, False))
        >>> state.deal_hole('7cTh')  # Carol*
        State.HoleDealing(player_index=1, cards=(7c, Th), statuses=(False, False))
        >>> state.deal_hole('Qc')  # Alice*
        State.HoleDealing(player_index=3, cards=(Qc,), statuses=(False,))

        >>> state.check_or_call()  # Bob
        State.CheckingOrCalling(player_index=0, amount=0)
        >>> state.complete_bet_or_raise_to()  # Carol
        State.CompletionBettingOrRaisingTo(player_index=1, amount=2)
        >>> state.check_or_call()  # Alice
        State.CheckingOrCalling(player_index=3, amount=2)
        >>> state.check_or_call()  # Bob
        State.CheckingOrCalling(player_index=0, amount=2)

        Below shows the second draw and actions.

        >>> state.stand_pat_or_discard('Js')  # Bob*
        State.StandingPatOrDiscarding(player_index=0, cards=(Js,))
        >>> state.stand_pat_or_discard()  # Carol*
        State.StandingPatOrDiscarding(player_index=1, cards=())
        >>> state.stand_pat_or_discard('Qc')  # Alice*
        State.StandingPatOrDiscarding(player_index=3, cards=(Qc,))
        >>> state.deal_hole('Ts')  # Bob*
        State.HoleDealing(player_index=0, cards=(Ts,), statuses=(False,))
        >>> state.deal_hole('9h')  # Alice*
        State.HoleDealing(player_index=3, cards=(9h,), statuses=(False,))

        >>> state.check_or_call()  # Bob
        State.CheckingOrCalling(player_index=0, amount=0)
        >>> state.complete_bet_or_raise_to()  # Carol
        State.CompletionBettingOrRaisingTo(player_index=1, amount=4)
        >>> state.complete_bet_or_raise_to()  # Alice
        State.CompletionBettingOrRaisingTo(player_index=3, amount=8)
        >>> state.fold()  # Bob
        State.Folding(player_index=0)
        >>> state.check_or_call()  # Carol
        State.CheckingOrCalling(player_index=1, amount=4)

        Below shows the third draw and actions.

        >>> state.stand_pat_or_discard('Th')  # Carol*
        State.StandingPatOrDiscarding(player_index=1, cards=(Th,))
        >>> state.stand_pat_or_discard()  # Alice*
        State.StandingPatOrDiscarding(player_index=3, cards=())
        >>> state.deal_hole('8h')  # Carol*
        State.HoleDealing(player_index=1, cards=(8h,), statuses=(False,))

        >>> state.check_or_call()  # Carol
        State.CheckingOrCalling(player_index=1, amount=0)
        >>> state.complete_bet_or_raise_to()  # Alice
        State.CompletionBettingOrRaisingTo(player_index=3, amount=4)
        >>> state.check_or_call()  # Carol
        State.CheckingOrCalling(player_index=1, amount=4)

        Below show the final stacks.

        >>> state.stacks
        [196, 220, 200, 184]

        :param antes: The antes.
        :param blinds_or_straddles: The blinds or straddles.
        :param small_bet: The small bet.
        :param big_bet: The big bet.
        :param starting_stacks: The starting stacks.
        :param player_count: The number of players.
        :return: The created state.
        """
        return State(
            Deck.REGULAR,
            (BadugiHand,),
            (
                Street(
                    False,
                    (False,) * 4,
                    0,
                    False,
                    Opening.POSITION,
                    small_bet,
                    None,
                ),
                Street(
                    True,
                    (),
                    0,
                    True,
                    Opening.POSITION,
                    small_bet,
                    None,
                ),
                Street(
                    True,
                    (),
                    0,
                    True,
                    Opening.POSITION,
                    big_bet,
                    None,
                ),
                Street(
                    True,
                    (),
                    0,
                    True,
                    Opening.POSITION,
                    big_bet,
                    None,
                ),
            ),
            BettingStructure.FIXED_LIMIT,
            automations,
            ante_trimming_status,
            cls._clean_values(antes, player_count),
            cls._clean_values(blinds_or_straddles, player_count),
            0,
            cls._clean_values(starting_stacks, player_count),
        )



""":mod:`pokerkit.hands` implements classes related to poker hands."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Hashable, Iterable, Iterator
from functools import total_ordering
from itertools import chain, combinations
from typing import Any

from pokerkit.lookups import (
    BadugiLookup,
    EightOrBetterLookup,
    Entry,
    KuhnPokerLookup,
    Lookup,
    RegularLowLookup,
    ShortDeckHoldemLookup,
    StandardLookup,
)
from pokerkit.utilities import Card, RankOrder


@total_ordering
class Hand(Hashable, ABC):
    """The abstract base class for poker hands.

    Stronger hands are considered greater than weaker hands.

    >>> h0 = ShortDeckHoldemHand(Card.parse('6s7s8s9sTs'))
    >>> h1 = ShortDeckHoldemHand(Card.parse('7c8c9cTcJc'))
    >>> h2 = ShortDeckHoldemHand(Card.parse('2c2d2h2s3h'))
    Traceback (most recent call last):
        ...
    ValueError: invalid hand '2c2d2h2s3h'
    >>> h0
    6s7s8s9sTs
    >>> h1
    7c8c9cTcJc
    >>> print(h0)
    Straight flush (6s7s8s9sTs)
    >>> h0 < h1
    True

    It does not make sense to compare hands of different types.

    >>> h = BadugiHand(Card.parse('6d7s8h9c'))
    >>> h < 500
    Traceback (most recent call last):
        ...
    TypeError: '<' not supported between instances of 'BadugiHand' and 'int'

    The hands are hashable.

    >>> h0 = ShortDeckHoldemHand(Card.parse('6s7s8s9sTs'))
    >>> h1 = ShortDeckHoldemHand(Card.parse('7c8c9cTcJc'))
    >>> hands = {h0, h1}
    """

    _lookup: Lookup
    _low: bool

    @classmethod
    @abstractmethod
    def from_game(
            cls,
            hole_cards: Iterable[Card],
            board_cards: Iterable[Card] = (),
    ) -> Hand:
        """Create a poker hand from a game setting.

        In a game setting, a player uses private cards from their hole
        and the public cards from the board to make their hand.

        :param hole_cards: The hole cards.
        :param board_cards: The optional board cards.
        :return: The strongest hand from possible card combinations.
        """

    def __init__(self, cards: Iterable[Card]) -> None:
        self.__cards = tuple(cards)

        if not self._lookup.has_entry(self.cards):
            raise ValueError(f'invalid hand \'{repr(self)}\'')

    def __eq__(self, other: Any) -> bool:
        if type(self) != type(other):
            return NotImplemented

        assert isinstance(other, Hand)

        return self.entry == other.entry

    def __hash__(self) -> int:
        return hash(self.entry)

    def __lt__(self, other: Hand) -> bool:
        if type(self) != type(other):
            return NotImplemented

        assert isinstance(other, Hand)

        if self._low:
            return self.entry > other.entry
        else:
            return self.entry < other.entry

    def __repr__(self) -> str:
        return ''.join(map(repr, self.cards))

    def __str__(self) -> str:
        return f'{self.entry.label.value} ({repr(self)})'

    @property
    def cards(self) -> tuple[Card, ...]:
        """Return the cards that form this hand.

        >>> hole = Card.parse('AsAc')
        >>> board = Card.parse('Kh3sAdAh')
        >>> hand = StandardHighHand.from_game(hole, board)
        >>> hand.cards
        (As, Ac, Kh, Ad, Ah)

        :return: The cards that form this hand.
        """
        return self.__cards

    @property
    def entry(self) -> Entry:
        """Return the hand entry.

        >>> hole = Card.parse('AsAc')
        >>> board = Card.parse('Kh3sAdAh')
        >>> hand = StandardHighHand.from_game(hole, board)
        >>> hand.entry.label
        <Label.FOUR_OF_A_KIND: 'Four of a kind'>

        :return: The hand entry.
        """
        return self._lookup.get_entry(self.cards)


class CombinationHand(Hand, ABC):
    """The abstract base class for combination hands."""

    _card_count: int

    @classmethod
    def from_game(
            cls,
            hole_cards: Iterable[Card],
            board_cards: Iterable[Card] = (),
    ) -> Hand:
        """Create a poker hand from a game setting.

        In a game setting, a player uses private cards from their hole
        and the public cards from the board to make their hand.

        >>> h0 = StandardHighHand.from_game(
        ...     Card.parse('AcAdAhAsKc'),
        ...     Card.parse(''),
        ... )
        >>> h1 = StandardHighHand(Card.parse('AcAdAhAsKc'))
        >>> h0 == h1
        True
        >>> h0 = StandardHighHand.from_game(
        ...     Card.parse('Ac9c'),
        ...     Card.parse('AhKhQhJhTh'),
        ... )
        >>> h1 = StandardHighHand(Card.parse('AhKhQhJhTh'))
        >>> h0 == h1
        True

        >>> h0 = StandardLowHand.from_game(
        ...     Card.parse('AcAdAhAsKc'),
        ...     Card.parse(''),
        ... )
        >>> h1 = StandardLowHand(Card.parse('AcAdAhAsKc'))
        >>> h0 == h1
        True
        >>> h0 = StandardLowHand.from_game(
        ...     Card.parse('Ac9c'),
        ...     Card.parse('AhKhQhJhTh'),
        ... )
        >>> h1 = StandardLowHand(Card.parse('AcQhJhTh9c'))
        >>> h0 == h1
        True

        >>> h0 = ShortDeckHoldemHand.from_game(
        ...     Card.parse('AcKs'),
        ...     Card.parse('AhAsKcJsTs'),
        ... )
        >>> h1 = ShortDeckHoldemHand(Card.parse('AcAhAsKcKs'))
        >>> h0 == h1
        True
        >>> h0 = ShortDeckHoldemHand.from_game(
        ...     Card.parse('AcAd'),
        ...     Card.parse('6s7cKcKd'),
        ... )
        >>> h1 = ShortDeckHoldemHand(Card.parse('AcAdKcKd7c'))
        >>> h0 == h1
        True

        >>> h0 = EightOrBetterLowHand.from_game(
        ...     Card.parse('As2s'),
        ...     Card.parse('2c3c4c5c6c'),
        ... )
        >>> h1 = EightOrBetterLowHand(Card.parse('Ad2d3d4d5d'))
        >>> h0 == h1
        True

        >>> h0 = RegularLowHand.from_game(
        ...     Card.parse('AcAd'),
        ...     Card.parse('AhAsKcQdQh'),
        ... )
        >>> h1 = RegularLowHand(Card.parse('AcAsQdQhKc'))
        >>> h0 == h1
        True
        >>> h0 = RegularLowHand.from_game(
        ...     Card.parse('AcAd'),
        ...     Card.parse('AhAsKcQd'),
        ... )
        >>> h1 = RegularLowHand(Card.parse('AdAhAsKcQd'))
        >>> h0 == h1
        True

        :param hole_cards: The hole cards.
        :param board_cards: The optional board cards.
        :return: The strongest hand from possible card combinations.
        """
        max_hand = None

        for combination in combinations(
                chain(hole_cards, board_cards),
                cls._card_count,
        ):
            try:
                hand = cls(combination)
            except ValueError:
                pass
            else:
                if max_hand is None or hand > max_hand:
                    max_hand = hand

        if max_hand is None:
            raise ValueError('no valid hand')

        return max_hand


class StandardHand(CombinationHand, ABC):
    """The abstract base class for standard hands."""

    _lookup = StandardLookup()
    _card_count = 5


class StandardHighHand(StandardHand):
    """The class for standard high hands.

    >>> h0 = StandardHighHand(Card.parse('7c5d4h3s2c'))
    >>> h1 = StandardHighHand(Card.parse('7c6d4h3s2c'))
    >>> h2 = StandardHighHand(Card.parse('8c7d6h4s2c'))
    >>> h3 = StandardHighHand(Card.parse('AcAsAd2s4s'))
    >>> h4 = StandardHighHand(Card.parse('TsJsQsKsAs'))
    >>> h0 < h1 < h2 < h3 < h4
    True

    >>> h = StandardHighHand(Card.parse('4c5dThJsAcKh2h'))
    Traceback (most recent call last):
        ...
    ValueError: invalid hand '4c5dThJsAcKh2h'
    >>> h = StandardHighHand(Card.parse('Ac2c3c4c'))
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'Ac2c3c4c'
    >>> h = StandardHighHand(())
    Traceback (most recent call last):
        ...
    ValueError: invalid hand ''
    """

    _low = False


class StandardLowHand(StandardHand):
    """The class for standard low hands.

    >>> h0 = StandardLowHand(Card.parse('TsJsQsKsAs'))
    >>> h1 = StandardLowHand(Card.parse('AcAsAd2s4s'))
    >>> h2 = StandardLowHand(Card.parse('8c7d6h4s2c'))
    >>> h3 = StandardLowHand(Card.parse('7c6d4h3s2c'))
    >>> h4 = StandardLowHand(Card.parse('7c5d4h3s2c'))
    >>> h0 < h1 < h2 < h3 < h4
    True

    >>> h = StandardLowHand(Card.parse('4c5dThJsAcKh2h'))
    Traceback (most recent call last):
        ...
    ValueError: invalid hand '4c5dThJsAcKh2h'
    >>> h = StandardLowHand(Card.parse('Ac2c3c4c'))
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'Ac2c3c4c'
    >>> h = StandardLowHand(())
    Traceback (most recent call last):
        ...
    ValueError: invalid hand ''
    """

    _low = True


class ShortDeckHoldemHand(CombinationHand):
    """The class for short-deck hold'em hands.

    Here, flushes beat full houses.

    >>> h0 = ShortDeckHoldemHand(Card.parse('6c7d8h9sJc'))
    >>> h1 = ShortDeckHoldemHand(Card.parse('7c7d7hTsQc'))
    >>> h2 = ShortDeckHoldemHand(Card.parse('As6c7h8h9h'))
    >>> h3 = ShortDeckHoldemHand(Card.parse('AsAhKcKhKd'))
    >>> h4 = ShortDeckHoldemHand(Card.parse('6s7s8sTsQs'))
    >>> h0 < h1 < h2 < h3 < h4
    True

    >>> h = ShortDeckHoldemHand(Card.parse('4c5dThJsAcKh2h'))
    Traceback (most recent call last):
        ...
    ValueError: invalid hand '4c5dThJsAcKh2h'
    >>> h = ShortDeckHoldemHand(Card.parse('Ac2c3c4c5c'))
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'Ac2c3c4c5c'
    >>> h = ShortDeckHoldemHand(())
    Traceback (most recent call last):
        ...
    ValueError: invalid hand ''
    """

    _lookup = ShortDeckHoldemLookup()
    _low = False
    _card_count = 5


class EightOrBetterLowHand(CombinationHand):
    """The class for eight or better low hands.

    >>> h0 = EightOrBetterLowHand(Card.parse('8c7d6h4s2c'))
    >>> h1 = EightOrBetterLowHand(Card.parse('7c5d4h3s2c'))
    >>> h2 = EightOrBetterLowHand(Card.parse('5d4h3s2dAd'))
    >>> h0 < h1 < h2
    True

    >>> h = EightOrBetterLowHand(Card.parse('AcAsAd2s4s'))
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'AcAsAd2s4s'
    >>> h = EightOrBetterLowHand(Card.parse('TsJsQsKsAs'))
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'TsJsQsKsAs'
    >>> h = EightOrBetterLowHand(Card.parse('4c5dThJsAcKh2h'))
    Traceback (most recent call last):
        ...
    ValueError: invalid hand '4c5dThJsAcKh2h'
    >>> h = EightOrBetterLowHand(Card.parse('Ac2c3c4c'))
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'Ac2c3c4c'
    >>> h = EightOrBetterLowHand(())
    Traceback (most recent call last):
        ...
    ValueError: invalid hand ''
    """

    _lookup = EightOrBetterLookup()
    _low = True
    _card_count = 5


class RegularLowHand(CombinationHand):
    """The class for low regular hands.

    Here, flushes are ignored.

    >>> h0 = RegularLowHand(Card.parse('KhKsKcKdAc'))
    >>> h1 = RegularLowHand(Card.parse('2s2c3s3cAh'))
    >>> h2 = RegularLowHand(Card.parse('6c4d3h2sAc'))
    >>> h3 = RegularLowHand(Card.parse('Ac2c3c4c5c'))
    >>> h0 < h1 < h2 < h3
    True

    >>> h = RegularLowHand(Card.parse('4c5dThJsAcKh2h'))
    Traceback (most recent call last):
        ...
    ValueError: invalid hand '4c5dThJsAcKh2h'
    >>> h = RegularLowHand(())
    Traceback (most recent call last):
        ...
    ValueError: invalid hand ''
    """

    _lookup = RegularLowLookup()
    _low = True
    _card_count = 5


class BoardCombinationHand(CombinationHand, ABC):
    """The abstract base class for board-combination hands."""

    _board_card_count: int

    @classmethod
    def from_game(
            cls,
            hole_cards: Iterable[Card],
            board_cards: Iterable[Card] = (),
    ) -> Hand:
        """Create a poker hand from a game setting.

        In a game setting, a player uses private cards from their hole
        and the public cards from the board to make their hand.

        >>> h0 = GreekHoldemHand.from_game(
        ...     Card.parse('Ac2d'),
        ...     Card.parse('QdJdTh2sKs'),
        ... )
        >>> h1 = GreekHoldemHand(Card.parse('2s2dAcKsQd'))
        >>> h0 == h1
        True
        >>> h0 = GreekHoldemHand.from_game(
        ...     Card.parse('AsKs'),
        ...     Card.parse('QdJdTh2s2d'),
        ... )
        >>> h1 = GreekHoldemHand(Card.parse('AsKsQdJdTh'))
        >>> h0 == h1
        True
        >>> h0 = GreekHoldemHand.from_game(
        ...     Card.parse('Ac9c'),
        ...     Card.parse('AhKhQhJhTh'),
        ... )
        >>> h1 = GreekHoldemHand(Card.parse('AcAhKhQh9c'))
        >>> h0 == h1
        True

        :param hole_cards: The hole cards.
        :param board_cards: The optional board cards.
        :return: The strongest hand from possible card combinations.
        """
        if isinstance(hole_cards, Iterator):
            hole_cards = tuple(hole_cards)

        max_hand = None

        for combination in combinations(board_cards, cls._board_card_count):
            try:
                hand = super().from_game(hole_cards, combination)
            except ValueError:
                pass
            else:
                if max_hand is None or hand > max_hand:
                    max_hand = hand

        if max_hand is None:
            raise ValueError('no valid hand')

        return max_hand


class GreekHoldemHand(BoardCombinationHand):
    """The class for Greek hold'em hands.

    In Greek hold'em, the player must use all of his or her hole cards
    to make a hand.

    >>> h0 = GreekHoldemHand(Card.parse('7c5d4h3s2c'))
    >>> h1 = GreekHoldemHand(Card.parse('7c6d4h3s2c'))
    >>> h2 = GreekHoldemHand(Card.parse('8c7d6h4s2c'))
    >>> h3 = GreekHoldemHand(Card.parse('AcAsAd2s4s'))
    >>> h4 = GreekHoldemHand(Card.parse('TsJsQsKsAs'))
    >>> h0 < h1 < h2 < h3 < h4
    True
    """

    _lookup = StandardLookup()
    _low = False
    _card_count = 5
    _board_card_count = 3


class HoleBoardCombinationHand(BoardCombinationHand, ABC):
    """The abstract base class for hole-board-combination hands."""

    _hole_card_count: int

    @classmethod
    def from_game(
            cls,
            hole_cards: Iterable[Card],
            board_cards: Iterable[Card] = (),
    ) -> Hand:
        """Create a poker hand from a game setting.

        In a game setting, a player uses private cards from their hole
        and the public cards from the board to make their hand.

        >>> h0 = OmahaHoldemHand.from_game(
        ...     Card.parse('6c7c8c9c'),
        ...     Card.parse('8s9sTc'),
        ... )
        >>> h1 = OmahaHoldemHand(Card.parse('6c7c8s9sTc'))
        >>> h0 == h1
        True
        >>> h0 = OmahaHoldemHand.from_game(
        ...     Card.parse('6c7c8s9s'),
        ...     Card.parse('8c9cTc'),
        ... )
        >>> h1 = OmahaHoldemHand(Card.parse('6c7c8c9cTc'))
        >>> h0 == h1
        True
        >>> h0 = OmahaHoldemHand.from_game(
        ...     Card.parse('6c7c8c9c'),
        ...     Card.parse('8s9sTc9hKs'),
        ... )
        >>> h1 = OmahaHoldemHand(Card.parse('8c8s9c9s9h'))
        >>> h0 == h1
        True
        >>> h0 = OmahaHoldemHand.from_game(
        ...     Card.parse('6c7c8sAh'),
        ...     Card.parse('As9cTc2sKs'),
        ... )
        >>> h1 = OmahaHoldemHand(Card.parse('AhAsKsTc8s'))
        >>> h0 == h1
        True

        >>> h0 = OmahaEightOrBetterLowHand.from_game(
        ...     Card.parse('As2s3s4s'),
        ...     Card.parse('2c3c4c5c6c'),
        ... )
        >>> h1 = OmahaEightOrBetterLowHand(Card.parse('Ad2d3d4d5d'))
        >>> h0 == h1
        True
        >>> h0 = OmahaEightOrBetterLowHand.from_game(
        ...     Card.parse('As6s7s8s'),
        ...     Card.parse('2c3c4c5c6c'),
        ... )
        >>> h1 = OmahaEightOrBetterLowHand(Card.parse('Ad2d3d4d6d'))
        >>> h0 == h1
        True

        :param hole_cards: The hole cards.
        :param board_cards: The optional board cards.
        :return: The strongest hand from possible card combinations.
        """
        if isinstance(board_cards, Iterator):
            board_cards = tuple(board_cards)

        max_hand = None

        for combination in combinations(hole_cards, cls._hole_card_count):
            try:
                hand = super().from_game(combination, board_cards)
            except ValueError:
                pass
            else:
                if max_hand is None or hand > max_hand:
                    max_hand = hand

        if max_hand is None:
            raise ValueError('no valid hand')

        return max_hand


class OmahaHoldemHand(HoleBoardCombinationHand):
    """The class for Omaha hold'em hands.

    In Omaha hold'em, the player must use a fixed number of his or her
    hole cards to make a hand.

    >>> h0 = OmahaHoldemHand(Card.parse('7c5d4h3s2c'))
    >>> h1 = OmahaHoldemHand(Card.parse('7c6d4h3s2c'))
    >>> h2 = OmahaHoldemHand(Card.parse('8c7d6h4s2c'))
    >>> h3 = OmahaHoldemHand(Card.parse('AcAsAd2s4s'))
    >>> h4 = OmahaHoldemHand(Card.parse('TsJsQsKsAs'))
    >>> h0 < h1 < h2 < h3 < h4
    True
    """

    _lookup = StandardLookup()
    _low = False
    _card_count = 5
    _board_card_count = 3
    _hole_card_count = 2


class OmahaEightOrBetterLowHand(HoleBoardCombinationHand):
    """The class for Omaha eight or better low hands.

    >>> h0 = OmahaEightOrBetterLowHand(Card.parse('8c7d6h4s2c'))
    >>> h1 = OmahaEightOrBetterLowHand(Card.parse('7c5d4h3s2c'))
    >>> h2 = OmahaEightOrBetterLowHand(Card.parse('5d4h3s2dAd'))
    >>> h0 < h1 < h2
    True
    """

    _lookup = EightOrBetterLookup()
    _low = True
    _card_count = 5
    _board_card_count = 3
    _hole_card_count = 2


class BadugiHand(Hand):
    """The class for badugi hands.

    >>> h0 = BadugiHand(Card.parse('Kc'))
    >>> h1 = BadugiHand(Card.parse('Ac'))
    >>> h2 = BadugiHand(Card.parse('4c8dKh'))
    >>> h3 = BadugiHand(Card.parse('Ac2d3h5s'))
    >>> h4 = BadugiHand(Card.parse('Ac2d3h4s'))
    >>> h0 < h1 < h2 < h3 < h4
    True

    >>> h = BadugiHand(Card.parse('Ac2d3c4s5c'))
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'Ac2d3c4s5c'
    >>> h = BadugiHand(Card.parse('Ac2d3c4s'))
    Traceback (most recent call last):
        ...
    ValueError: cards not rainbow
    >>> h = BadugiHand(Card.parse('AcAd3h4s'))
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'AcAd3h4s'
    >>> h = BadugiHand(Card.parse('Ac2c'))
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'Ac2c'
    >>> h = BadugiHand(())
    Traceback (most recent call last):
        ...
    ValueError: invalid hand ''
    """

    _lookup = BadugiLookup()
    _low = True

    @classmethod
    def from_game(
            cls,
            hole_cards: Iterable[Card],
            board_cards: Iterable[Card] = (),
    ) -> Hand:
        """Create a poker hand from a game setting.

        In a game setting, a player uses private cards from their hole
        and the public cards from the board to make their hand.

        >>> h0 = BadugiHand.from_game(Card.parse('2s4c5d6h'))
        >>> h1 = BadugiHand(Card.parse('2s4c5d6h'))
        >>> h0 == h1
        True
        >>> h0 = BadugiHand.from_game(Card.parse('2s3s4d7h'))
        >>> h1 = BadugiHand(Card.parse('2s4d7h'))
        >>> h0 == h1
        True
        >>> h0 = BadugiHand.from_game(Card.parse('KcKdKhKs'))
        >>> h1 = BadugiHand(Card.parse('Ks'))
        >>> h0 == h1
        True

        :param hole_cards: The hole cards.
        :param board_cards: The optional board cards.
        :return: The strongest hand from possible card combinations.
        """
        ranks = set()
        suits = set()
        cards = set()

        for card in sorted(
                chain(hole_cards, board_cards),
                key=lambda card: RankOrder.STANDARD.index(card.rank),
        ):
            if card.rank not in ranks and card.suit not in suits:
                ranks.add(card.rank)
                suits.add(card.suit)
                cards.add(card)

        return cls(cards)

    def __init__(self, cards: Iterable[Card]) -> None:
        super().__init__(cards)

        if not Card.are_rainbow(self.cards):
            raise ValueError('cards not rainbow')


class KuhnPokerHand(Hand):
    """The class for Kuhn poker hands.

    >>> h0 = KuhnPokerHand(Card.parse('Js'))
    >>> h1 = KuhnPokerHand(Card.parse('Qs'))
    >>> h2 = KuhnPokerHand(Card.parse('Ks'))
    >>> h0 < h1 < h2
    True

    >>> h = KuhnPokerHand(Card.parse('As'))
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'As'
    """

    _lookup = KuhnPokerLookup()
    _low = False

    @classmethod
    def from_game(
            cls,
            hole_cards: Iterable[Card],
            board_cards: Iterable[Card] = (),
    ) -> Hand:
        """Create a poker hand from a game setting.

        In a game setting, a player uses private cards from their hole
        and the public cards from the board to make their hand.

        >>> h0 = BadugiHand.from_game(Card.parse('Ks'))
        >>> h1 = BadugiHand(Card.parse('Ks'))
        >>> h0 == h1
        True

        :param hole_cards: The hole cards.
        :param board_cards: The optional board cards.
        :return: The strongest hand from possible card combinations.
        """
        return max(map(cls, chain(hole_cards, board_cards)))



""":mod:`pokerkit.tests.test_games` implements unit tests for
:mod:`pokerkit.test_games`.
"""

from unittest import TestCase

from pokerkit.games import Poker


class StandardLookupTestCase(TestCase):
    def test_clean_values(self) -> None:
        self.assertEqual(Poker._clean_values(5, 6), (5,) * 6)
        self.assertEqual(Poker._clean_values([1, 2], 6), (1, 2, 0, 0, 0, 0))
        self.assertEqual(
            Poker._clean_values({0: 1, 1: 2}, 6),
            (1, 2, 0, 0, 0, 0),
        )
        self.assertEqual(Poker._clean_values({-1: 2}, 6), (0, 0, 0, 0, 0, 2))
        self.assertEqual(Poker._clean_values(None, 6), (0, 0, 0, 0, 0, 0))
        self.assertRaises(AssertionError, Poker._clean_values, type, 6)



""":mod:`pokerkit.utilities` implements classes related to poker
utilities.
"""

from __future__ import annotations

from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from enum import Enum, StrEnum, unique
from functools import partial
from itertools import product, starmap
from operator import is_not
from typing import Any


@unique
class Rank(StrEnum):
    """The enum class for ranks.

    A card of lower rank is said to be less than that of a higher rank.
    The ordering of the ranks is decided in accordance to the rules of
    the poker variant being played.

    >>> Rank.DEUCE
    <Rank.DEUCE: '2'>
    >>> Rank('2')
    <Rank.DEUCE: '2'>
    >>> Rank.DEUCE.value
    '2'
    >>> Rank.DEUCE.name
    'DEUCE'
    """

    ACE: str = 'A'
    """The rank of aces."""
    DEUCE: str = '2'
    """The rank of deuces."""
    TREY: str = '3'
    """The rank of treys."""
    FOUR: str = '4'
    """The rank of fours."""
    FIVE: str = '5'
    """The rank of fives."""
    SIX: str = '6'
    """The rank of sixes."""
    SEVEN: str = '7'
    """The rank of sevens."""
    EIGHT: str = '8'
    """The rank of eights."""
    NINE: str = '9'
    """The rank of nines."""
    TEN: str = 'T'
    """The rank of tens."""
    JACK: str = 'J'
    """The rank of Jacks."""
    QUEEN: str = 'Q'
    """The rank of Queens."""
    KING: str = 'K'
    """The rank of Kings."""


@unique
class RankOrder(tuple[Rank, ...], Enum):
    """The enum class for tuples of ranks.

    Poker variants order the ranks differently between each other. Most
    games use deuce to ace ordering which is defined here as
    :attr:`RankOrder.STANDARD`.

    >>> len(RankOrder.STANDARD)
    13
    >>> RankOrder.STANDARD[0]
    <Rank.DEUCE: '2'>
    >>> RankOrder.STANDARD[-1]
    <Rank.ACE: 'A'>

    Short-deck hold'em uses its own ranking which eliminates deuces,
    treys, fours, and fives, defined here as
    :attr:`RankOrder.SHORT_DECK_HOLDEM`.

    >>> len(RankOrder.SHORT_DECK_HOLDEM)
    9
    >>> RankOrder.SHORT_DECK_HOLDEM[0]
    <Rank.SIX: '6'>
    >>> RankOrder.SHORT_DECK_HOLDEM[-1]
    <Rank.ACE: 'A'>

    In some draw games, Ace is considered to be the lowest rank. This
    ordering is accessible via :attr:`RankOrder.REGULAR`.

    >>> len(RankOrder.REGULAR)
    13
    >>> RankOrder.REGULAR[0]
    <Rank.ACE: 'A'>
    >>> RankOrder.REGULAR[-1]
    <Rank.KING: 'K'>
    """

    STANDARD: tuple[Rank, ...] = (
        Rank.DEUCE,
        Rank.TREY,
        Rank.FOUR,
        Rank.FIVE,
        Rank.SIX,
        Rank.SEVEN,
        Rank.EIGHT,
        Rank.NINE,
        Rank.TEN,
        Rank.JACK,
        Rank.QUEEN,
        Rank.KING,
        Rank.ACE,
    )
    """The standard ranks (from deuce to ace)."""
    SHORT_DECK_HOLDEM: tuple[Rank, ...] = (
        Rank.SIX,
        Rank.SEVEN,
        Rank.EIGHT,
        Rank.NINE,
        Rank.TEN,
        Rank.JACK,
        Rank.QUEEN,
        Rank.KING,
        Rank.ACE,
    )
    """The short-deck hold'em ranks (from six to ace)."""
    REGULAR: tuple[Rank, ...] = (
        Rank.ACE,
        Rank.DEUCE,
        Rank.TREY,
        Rank.FOUR,
        Rank.FIVE,
        Rank.SIX,
        Rank.SEVEN,
        Rank.EIGHT,
        Rank.NINE,
        Rank.TEN,
        Rank.JACK,
        Rank.QUEEN,
        Rank.KING,
    )
    """The regular ranks (from ace to king)."""
    EIGHT_OR_BETTER_LOW: tuple[Rank, ...] = (
        Rank.ACE,
        Rank.DEUCE,
        Rank.TREY,
        Rank.FOUR,
        Rank.FIVE,
        Rank.SIX,
        Rank.SEVEN,
        Rank.EIGHT,
    )
    """The eight or better low ranks (from ace to eight)."""
    KUHN_POKER: tuple[Rank, ...] = Rank.JACK, Rank.QUEEN, Rank.KING
    """The Kuhn poker ranks (from jack to king)."""


@unique
class Suit(StrEnum):
    """The enum class for suits.

    As in the case of every poker games, the suit's ordering is clubs,
    diamonds, hearts, and spades, which is also alphabetical.

    In most poker variants, suits are completely irrelevant and are
    never used to break ties or to decide the winner.

    In stud games, suits are used to break ties when deciding the
    opener when there are multiple up cards with the same rank.

    >>> Suit.CLUB
    <Suit.CLUB: 'c'>
    >>> Suit('c')
    <Suit.CLUB: 'c'>
    >>> Suit.CLUB.name
    'CLUB'
    >>> Suit.CLUB.value
    'c'
    """

    CLUB: str = 'c'
    """The suit of clubs."""
    DIAMOND: str = 'd'
    """The suit of diamonds."""
    HEART: str = 'h'
    """The suit of hearts."""
    SPADE: str = 's'
    """The suit of spades."""


@dataclass(frozen=True)
class Card:
    """The class for cards.

    Cards are described by their ranks and their suits.

    Cards with higher ranks are considered stronger than those with
    lower ranks. Of course, the rank ordering is variant-dependant.
    Ties are broken by suits in some rare variants.

    >>> card = Card(Rank.ACE, Suit.SPADE)
    >>> card
    As
    >>> str(card)
    'ACE OF SPADES (As)'
    >>> card.rank
    <Rank.ACE: 'A'>
    >>> card.suit
    <Suit.SPADE: 's'>

    Note that the card attributes are read-only.

    >>> card.rank = Rank.KING
    Traceback (most recent call last):
        ...
    dataclasses.FrozenInstanceError: cannot assign to field 'rank'

    The card is also hashable.

    >>> from collections.abc import Hashable
    >>> isinstance(card, Hashable)
    True
    """

    rank: Rank
    """The rank."""
    suit: Suit
    """The suit."""

    @classmethod
    def get_ranks(cls, cards: Iterable[Card]) -> Iterator[Rank]:
        """Return an iterator of the ranks of each card.

        >>> Card.get_ranks(Card.parse('2sKh'))  # doctest: +ELLIPSIS
        <generator object Card.get_ranks at 0x...>
        >>> list(Card.get_ranks(Card.parse('2sKh')))
        [<Rank.DEUCE: '2'>, <Rank.KING: 'K'>]

        :param cards: The cards to get ranks from.
        :return: The iterator of the ranks of each card.
        """
        for card in cards:
            yield card.rank

    @classmethod
    def get_suits(cls, cards: Iterable[Card]) -> Iterator[Suit]:
        """Return an iterator of the suits of each card.

        >>> Card.get_suits(Card.parse('2sKh'))  # doctest: +ELLIPSIS
        <generator object Card.get_suits at 0x...>
        >>> list(Card.get_suits(Card.parse('2sKh')))
        [<Suit.SPADE: 's'>, <Suit.HEART: 'h'>]

        :param cards: The cards to get suits from.
        :return: The iterator of the suits of each card.
        """
        for card in cards:
            yield card.suit

    @classmethod
    def are_paired(cls, cards: Iterable[Card]) -> bool:
        """Return the pairedness of the given cards.

        The cards are paired if at least two of the cards share a
        common rank.

        >>> Card.are_paired(())
        False
        >>> Card.are_paired(Card.parse('2sKh'))
        False
        >>> Card.are_paired(Card.parse('2sKh2h'))
        True
        >>> Card.are_paired(Card.parse('2s2c2h'))
        True

        :param cards: The cards to determine the pairedness from.
        :return: The pairedness of the cards.
        """
        ranks = tuple(cls.get_ranks(cards))

        return len(set(ranks)) != len(ranks)

    @classmethod
    def are_suited(cls, cards: Iterable[Card]) -> bool:
        """Return the suitedness of the given cards.

        The cards are suited if all of the cards share a common suit.

        >>> Card.are_suited(())
        True
        >>> Card.are_suited(Card.parse('2s'))
        True
        >>> Card.are_suited(Card.parse('2sKh3s'))
        False
        >>> Card.are_suited(Card.parse('2hKh3h'))
        True

        :param cards: The cards to determine the suitedness from.
        :return: The suitedness of the cards.
        """
        return len(set(cls.get_suits(cards))) <= 1

    @classmethod
    def are_rainbow(cls, cards: Iterable[Card]) -> bool:
        """Return whether the cards are rainbow.

        The cards are rainbow if no two cards share a common suit.

        >>> Card.are_rainbow(())
        True
        >>> Card.are_rainbow(Card.parse('2s'))
        True
        >>> Card.are_rainbow(Card.parse('2sKh3c'))
        True
        >>> Card.are_rainbow(Card.parse('2hKh3c'))
        False

        :param cards: The cards to determine whether they are rainbow.
        :return: ``True`` if the cards are rainbow, otherwise ``False``.
        """
        suits = tuple(cls.get_suits(cards))

        return len(set(suits)) == len(suits)

    @classmethod
    def parse(cls, *raw_cards: str) -> Iterator[Card]:
        """Parse the string of the card representations.

        >>> Card.parse('AsKsQsJsTs')  # doctest: +ELLIPSIS
        <generator object Card.parse at 0x...>
        >>> list(Card.parse('2c8d5sKh'))
        [2c, 8d, 5s, Kh]
        >>> next(Card.parse('AcAh'))
        Ac
        >>> next(Card.parse('AcA'))
        Traceback (most recent call last):
            ...
        ValueError: content length not a multiple of 2
        >>> next(Card.parse('1d'))
        Traceback (most recent call last):
            ...
        ValueError: '1' is not a valid Rank
        >>> next(Card.parse('Ax'))
        Traceback (most recent call last):
            ...
        ValueError: 'x' is not a valid Suit

        :param raw_cards: The string of the card representations.
        :return: The generator yielding the parsed cards.
        :raises ValueError: If any card representation is invalid.
        """
        for contents in raw_cards:
            for content in contents.split():
                if len(content) % 2 != 0:
                    raise ValueError('content length not a multiple of 2')

                for i in range(0, len(content), 2):
                    rank = Rank(content[i])
                    suit = Suit(content[i + 1])

                    yield cls(rank, suit)

    def __repr__(self) -> str:
        return f'{self.rank}{self.suit}'

    def __str__(self) -> str:
        return f'{self.rank.name} OF {self.suit.name}S ({repr(self)})'


@unique
class Deck(tuple[Card, ...], Enum):
    """The enum class for a tuple of cards representing decks.

    Most poker games use the 52-card standard deck, denoted as
    :attr:`Deck.STANDARD`.

    >>> len(Deck.STANDARD)
    52
    >>> Deck.STANDARD[:6]
    (2c, 2d, 2h, 2s, 3c, 3d)

    However, short-deck hold'em uses a 36-card short deck where cards
    with ranks from 2 to 5 (inclusive) are missing. This particular deck
    is accessible via :attr:`Deck.SHORT_DECK_HOLDEM`.

    >>> len(Deck.SHORT_DECK_HOLDEM)
    36
    >>> Deck.SHORT_DECK_HOLDEM[:6]
    (6c, 6d, 6h, 6s, 7c, 7d)

    The members of :class:`Deck` are tuples and cannot be modified. To
    use it effectively, you must make a copy with a different data type
    like ``list`` or ``deque``.

    >>> Deck.STANDARD[3] = Card(Rank.ACE, Suit.SPADE)
    Traceback (most recent call last):
        ...
    TypeError: 'Deck' object does not support item assignment
    >>> deck = list(Deck.STANDARD)
    >>> deck[3] = Card(Rank.ACE, Suit.SPADE)
    >>> deck[:6]
    [2c, 2d, 2h, As, 3c, 3d]

    The standard and regular decks have identical contents. The ordering
    of the cards are different, however.

    >>> len(Deck.STANDARD)
    52
    >>> len(Deck.REGULAR)
    52
    >>> Deck.STANDARD[:6]
    (2c, 2d, 2h, 2s, 3c, 3d)
    >>> Deck.REGULAR[:6]
    (Ac, Ad, Ah, As, 2c, 2d)
    """

    STANDARD: tuple[Card, ...] = tuple(
        starmap(
            Card,
            product(
                RankOrder.STANDARD,
                (Suit.CLUB, Suit.DIAMOND, Suit.HEART, Suit.SPADE),
            ),
        ),
    )
    """The 52-card standard deck cards.

    The ranks are ordered from deuces to aces.
    """
    SHORT_DECK_HOLDEM: tuple[Card, ...] = tuple(
        starmap(
            Card,
            product(
                RankOrder.SHORT_DECK_HOLDEM,
                (Suit.CLUB, Suit.DIAMOND, Suit.HEART, Suit.SPADE),
            ),
        ),
    )
    """The 36-card short-deck hold'em cards.

    In this deck, deuces to fives are stripped away.
    """
    REGULAR: tuple[Card, ...] = tuple(
        starmap(
            Card,
            product(
                RankOrder.REGULAR,
                (Suit.CLUB, Suit.DIAMOND, Suit.HEART, Suit.SPADE),
            ),
        ),
    )
    """The 52-card regular deck cards.

    The ranks are ordered from aces to kings.
    """
    KUHN_POKER: tuple[Card, ...] = tuple(Card.parse('JsQsKs'))
    """The 3-card Kuhn poker deck cards.

    The cards in it are jack of spades, queen of spades, and king of
    spades.
    """


def filter_none(values: Iterable[Any]) -> Any:
    """Filter ``None`` from values.

    >>> filter_none([1, 2, None, 3])  # doctest: +ELLIPSIS
    <filter object at 0x...>
    >>> list(filter_none([1, 2, None, 3]))
    [1, 2, 3]

    :param values: The optional values.
    :return: The filtered values.
    """
    return filter(partial(is_not, None), values)


def min_or_none(values: Iterable[Any], key: Any = None) -> Any:
    """Get the minimum value while ignoring ``None`` values.

    >>> min_or_none([1, 2, 3, -2, -1])
    -2
    >>> min_or_none([1, None, 2, None, 3, -2, -1, None])
    -2
    >>> min_or_none([]) is None
    True

    :param values: The optional values.
    :param key: The optional key function.
    :return: The minimum value if any, otherwise ``None``.
    """
    try:
        return min(filter_none(values), key=key)
    except ValueError:
        return None


def max_or_none(values: Iterable[Any], key: Any = None) -> Any:
    """Get the maximum value while ignoring ``None`` values.

    >>> max_or_none([1, 2, 3, -2, -1])
    3
    >>> max_or_none([1, None, 2, None, 3, -2, -1, None])
    3
    >>> max_or_none([]) is None
    True

    :param values: The optional values.
    :param key: The optional key function.
    :return: The maximum value if any, otherwise ``None``.
    """
    try:
        return max(filter_none(values), key=key)
    except ValueError:
        return None



""":mod:`pokerkit.tests.test_lookups` implements unit tests for
:mod:`pokerkit.test_lookups`.
"""

from collections.abc import Iterable
from hashlib import md5
from itertools import combinations
from unittest import TestCase

from pokerkit.lookups import (
    BadugiLookup,
    EightOrBetterLookup,
    RegularLowLookup,
    ShortDeckHoldemLookup,
    StandardLookup,
)
from pokerkit.utilities import Card, Deck


class LookupTestCaseMixin:
    @classmethod
    def serialize_combinations(
            cls,
            combinations_: Iterable[Iterable[Card]],
    ) -> str:
        return '\n'.join(map(cls.serialize_combination, combinations_))

    @classmethod
    def serialize_combination(cls, combination: Iterable[Card]) -> str:
        return ''.join(map(repr, combination))


class StandardLookupTestCase(LookupTestCaseMixin, TestCase):
    def test_get_entry(self) -> None:
        lookup = StandardLookup()
        combinations_ = sorted(
            combinations(Deck.STANDARD, 5),
            key=lookup.get_entry,
        )
        string = self.serialize_combinations(combinations_)
        algorithm = md5()
        algorithm.update(string.encode())

        self.assertEqual(
            algorithm.hexdigest(),
            '488cdd27873395ba75205cd02fb9d6b2',
        )


class ShortDeckHoldemLookupTestCase(LookupTestCaseMixin, TestCase):
    def test_get_entry(self) -> None:
        lookup = ShortDeckHoldemLookup()
        combinations_ = sorted(
            combinations(Deck.SHORT_DECK_HOLDEM, 5),
            key=lookup.get_entry,
        )
        string = self.serialize_combinations(combinations_)
        algorithm = md5()
        algorithm.update(string.encode())

        self.assertEqual(
            algorithm.hexdigest(),
            '5b46b727ce4526a68d41b0e55939353c',
        )


class EightOrBetterLookupTestCase(LookupTestCaseMixin, TestCase):
    def test_get_entry(self) -> None:
        lookup = EightOrBetterLookup()
        combinations_ = sorted(
            filter(lookup.has_entry, combinations(Deck.STANDARD, 5)),
            key=lookup.get_entry,
        )
        string = self.serialize_combinations(combinations_)
        algorithm = md5()
        algorithm.update(string.encode())

        self.assertEqual(
            algorithm.hexdigest(),
            'd765d50ed799d03a8f062a3d5a31ad5c',
        )


class RegularLowLookupTestCase(LookupTestCaseMixin, TestCase):
    def test_get_entry(self) -> None:
        lookup = RegularLowLookup()
        combinations_ = sorted(
            combinations(Deck.STANDARD, 5),
            key=lookup.get_entry,
        )
        string = self.serialize_combinations(combinations_)
        algorithm = md5()
        algorithm.update(string.encode())

        self.assertEqual(
            algorithm.hexdigest(),
            'edd288a3022b2763475a533ad2a102e9',
        )


class BadugiLookupTestCase(LookupTestCaseMixin, TestCase):
    def test_get_entry(self) -> None:
        lookup = BadugiLookup()
        combinations_ = []

        for n in range(1, 5):
            for cards in combinations(Deck.STANDARD, n):
                pairedness = Card.are_paired(cards)
                suitedness = Card.are_suited(cards)
                rainbowness = Card.are_rainbow(cards)

                if not pairedness and rainbowness:
                    combinations_.append(cards)
                elif pairedness or suitedness:
                    self.assertRaises(ValueError, lookup.get_entry, cards)

        combinations_.sort(key=lookup.get_entry)
        string = self.serialize_combinations(combinations_)
        algorithm = md5()
        algorithm.update(string.encode())

        self.assertEqual(
            algorithm.hexdigest(),
            'c284d3bf45c87707f62cd7fc6699c31a',
        )



#!/usr/bin/env python3

from setuptools import find_packages, setup

with open('README.rst', 'r') as file:
    long_description = file.read()

setup(
    name='pokerkit',
    version='0.0.0',
    description='A Python package for various poker tools',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/uoftcprg/pokerkit',
    author='University of Toronto Computer Poker Research Group',
    author_email='uoftcprg@outlook.com',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Topic :: Education',
        'Topic :: Games/Entertainment',
        'Topic :: Games/Entertainment :: Board Games',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    keywords=[
        'poker',
        'nlhe',
        'ai',
        'game',
        'game theory',
        'libratus',
        'modicum',
    ],
    project_urls={
        'Documentation': 'https://pokerkit.readthedocs.io/en/latest/',
        'Source': 'https://github.com/uoftcprg/pokerkit',
        'Tracker': 'https://github.com/uoftcprg/pokerkit/issues',
    },
    package_data={'pokerkit': ['py.typed']},
    packages=find_packages(),
    python_requires='>=3.11',
)
